from amltk import options
from amltk.optimization import (
    History,
    Metric,
    Optimizer,
    Trial,
)
from amltk.pipeline import (
    Choice,
    Component,
    Fixed,
    Join,
    Node,
    Sequential,
    Split,
    request,
)
from amltk.scheduling import (
    Comm,
    Emitter,
    Event,
    Limiter,
    Plugin,
    Scheduler,
    SequentialExecutor,
    Subscriber,
    Task,
)
from amltk.store import (
    Bucket,
    ByteLoader,
    Drop,
    JSONLoader,
    Loader,
    NPYLoader,
    PathBucket,
    PathLoader,
    PDLoader,
    PickleLoader,
    TxtLoader,
    YAMLLoader,
)

__all__ = [
    "Bucket",
    "ByteLoader",
    "Choice",
    "Comm",
    "Component",
    "Drop",
    "Emitter",
    "Event",
    "Fixed",
    "History",
    "Join",
    "Metric",
    "JSONLoader",
    "Limiter",
    "Loader",
    "Node",
    "NPYLoader",
    "Optimizer",
    "options",
    "PathBucket",
    "PathLoader",
    "PDLoader",
    "PickleLoader",
    "Plugin",
    "request",
    "Scheduler",
    "Scheduler",
    "Sequential",
    "SequentialExecutor",
    "Split",
    "Subscriber",
    "Task",
    "Trial",
    "TxtLoader",
    "YAMLLoader",
]



"""A module holding a decorator to wrap a function to add a traceback to
any exception raised.
"""
from __future__ import annotations

import traceback
from collections.abc import Callable, Iterable, Iterator
from typing import TYPE_CHECKING, Any, TypeVar
from typing_extensions import ParamSpec, override

if TYPE_CHECKING:
    from amltk.pipeline.node import Node

R = TypeVar("R")
E = TypeVar("E")
P = ParamSpec("P")


def safe_map(
    f: Callable[..., R],
    args: Iterable[Any],
) -> Iterator[R | tuple[Exception, str]]:
    """Map a function over an iterable, catching any exceptions.

    Args:
        f: The function to map.
        args: The iterable to map over.

    Yields:
        The return value of the function, or the exception raised.
    """
    for arg in args:
        try:
            yield f(arg)
        except Exception as e:  # noqa: BLE001
            yield e, traceback.format_exc()


def safe_starmap(
    f: Callable[..., R],
    args: Iterable[Iterable[Any]],
) -> Iterator[R | tuple[Exception, str]]:
    """Map a function over an iterable, catching any exceptions.

    Args:
        f: The function to map.
        args: The iterable to map over.

    Yields:
        The return value of the function, or the exception raised.
    """
    for arg in args:
        try:
            yield f(*arg)
        except Exception as e:  # noqa: BLE001
            yield e, traceback.format_exc()


class IntegrationNotFoundError(Exception):
    """An exception raised when no integration is found."""

    def __init__(self, name: str) -> None:
        """Initialize the exception.

        Args:
            name: The name of the integration that was not found.
        """
        super().__init__(f"No integration found for {name}.")


class SchedulerNotRunningError(RuntimeError):
    """The scheduler is not running."""


class EventNotKnownError(ValueError):
    """The event is not a known one."""


class NoChoiceMadeError(ValueError):
    """No choice was made."""


class NodeNotFoundError(ValueError):
    """The node was not found."""


class RequestNotMetError(ValueError):
    """Raised when a request is not met."""


class DuplicateNamesError(ValueError):
    """Raised when duplicate names are found."""

    def __init__(self, node: Node) -> None:
        """Initialize the exception.

        Args:
            node: The node that has children with duplicate names.
        """
        super().__init__(node)
        self.node = node

    @override
    def __str__(self) -> str:
        return (
            f"Duplicate names found in {self.node.name} and can't be handled."
            f"\nnodes: {[n.name for n in self.node.nodes]}."
        )



"""The [`SMACOptimizer`][amltk.optimization.optimizers.smac.SMACOptimizer],
is a wrapper around the [`smac`](https://github.com/automl/smac3) optimizer.

!!! tip "Requirements"

    This requires `smac` which can be installed with:

    ```bash
    pip install amltk[smac]

    # Or directly
    pip install smac
    ```

This uses `ConfigSpace` as its [`search_space()`][amltk.pipeline.Node.search_space] to
optimize.

Users should report results using
[`trial.success()`][amltk.optimization.Trial.success].

Visit their documentation for what you can pass to
[`SMACOptimizer.create()`][amltk.optimization.optimizers.smac.SMACOptimizer.create].

The below example shows how you can use SMAC to optimize an sklearn pipeline.

```python exec="True" source="material-block" result="python"
from __future__ import annotations

import logging

from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split

from amltk.optimization.optimizers.smac import SMACOptimizer
from amltk.scheduling import Scheduler
from amltk.optimization import History, Trial
from amltk.pipeline import Component, Node

logging.basicConfig(level=logging.INFO)


def target_function(trial: Trial, pipeline: Node) -> Trial.Report:
    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y)
    clf = pipeline.configure(trial.config).build("sklearn")

    with trial.begin():
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)
        accuracy = accuracy_score(y_test, y_pred)
        loss = 1 - accuracy
        return trial.success(loss=loss, accuracy=accuracy)

    return trial.fail()
from amltk._doc import make_picklable; make_picklable(target_function)  # markdown-exec: hide


pipeline = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
space = pipeline.search_space(parser=SMACOptimizer.preferred_parser())
optimizer = SMACOptimizer.create(space=space)

N_WORKERS = 2
scheduler = Scheduler.with_processes(N_WORKERS)
task = scheduler.task(target_function)

history = History()

@scheduler.on_start(repeat=N_WORKERS)
def on_start():
    trial = optimizer.ask()
    task.submit(trial, pipeline)

@task.on_result
def tell_and_launch_trial(_, report: Trial.Report):
    if scheduler.running():
        optimizer.tell(report)
        trial = optimizer.ask()
        task.submit(trial, pipeline)

@task.on_result
def add_to_history(_, report: Trial.Report):
    history.add(report)

scheduler.run(timeout=3, wait=False)

print(history.df())
"""  # noqa: E501
from __future__ import annotations

import logging
from collections.abc import Mapping, Sequence
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, overload
from typing_extensions import override

import numpy as np
from more_itertools import first_true
from pynisher import MemoryLimitException, TimeoutException
from smac import HyperparameterOptimizationFacade, MultiFidelityFacade, Scenario
from smac.runhistory import (
    StatusType,
    TrialInfo as SMACTrialInfo,
    TrialValue as SMACTrialValue,
)

from amltk.optimization import Metric, Optimizer, Trial
from amltk.pipeline import Node
from amltk.randomness import as_int
from amltk.store import PathBucket

if TYPE_CHECKING:
    from typing_extensions import Self

    from ConfigSpace import ConfigurationSpace
    from smac.facade import AbstractFacade

    from amltk.types import FidT, Seed


logger = logging.getLogger(__name__)


# NOTE: In this code, we refer to the distance to the optimal as
# cost. This is because SMAC deals with minization and has called
# it cost. Cost can have other meaning in the context of different
# optimizers so we localize the meaning of cost here.


class SMACOptimizer(Optimizer[SMACTrialInfo]):
    """An optimizer that uses SMAC to optimize a config space."""

    def __init__(
        self,
        *,
        facade: AbstractFacade,
        bucket: PathBucket | None = None,
        metrics: Metric | Sequence[Metric],
        fidelities: Mapping[str, FidT] | None = None,
    ) -> None:
        """Initialize the optimizer.

        Args:
            facade: The SMAC facade to use.
            bucket: The bucket given to trials generated by this optimizer.
            metrics: The metrics to optimize.
            fidelities: The fidelities to use, if any.
        """
        # We need to very that the scenario is correct incase user pass in
        # their own facade construction
        assert self.crash_cost(metrics) == facade.scenario.crash_cost

        metrics = metrics if isinstance(metrics, Sequence) else [metrics]
        super().__init__(metrics=metrics, bucket=bucket)
        self.facade = facade
        self.metrics = metrics
        self.fidelities = fidelities

    @classmethod
    def create(
        cls,
        *,
        space: ConfigurationSpace | Node,
        metrics: Metric | Sequence[Metric],
        bucket: PathBucket | str | Path | None = None,
        deterministic: bool = True,
        seed: Seed | None = None,
        fidelities: Mapping[str, FidT] | None = None,
        continue_from_last_run: bool = False,
        logging_level: int | Path | Literal[False] | None = False,
    ) -> Self:
        """Create a new SMAC optimizer using either the HPO facade or
        a mutli-fidelity facade.

        Args:
            space: The config space to optimize.
            metrics: The metrics to optimize.
            bucket: The bucket given to trials generated by this optimizer.
            deterministic: Whether the function your optimizing is deterministic, given
                a seed and config.
            seed: The seed to use for the optimizer.
            fidelities: The fidelities to use, if any.
            continue_from_last_run: Whether to continue from a previous run.
            logging_level: The logging level to use.
                This argument is passed forward to SMAC, use False to disable
                SMAC's handling of logging.
        """
        seed = as_int(seed)
        match bucket:
            case None:
                bucket = PathBucket(
                    f"{cls.__class__.__name__}-{datetime.now().isoformat()}",
                )
            case str() | Path():
                bucket = PathBucket(bucket)
            case bucket:
                bucket = bucket  # noqa: PLW0127

        # NOTE SMAC always minimizes! Hence we make it a minimization problem
        metric_names: str | list[str]
        if isinstance(metrics, Sequence):
            metric_names = [metric.name for metric in metrics]
        else:
            metric_names = metrics.name

        if isinstance(space, Node):
            space = space.search_space(parser=cls.preferred_parser())

        facade_cls: type[AbstractFacade]
        if fidelities:
            if len(fidelities) == 1:
                v = next(iter(fidelities.values()))
                min_budget, max_budget = v
            else:
                min_budget, max_budget = 1.0, 100.0

            scenario = Scenario(
                objectives=metric_names,
                configspace=space,
                output_directory=bucket.path / "smac3_output",
                seed=seed,
                min_budget=min_budget,
                max_budget=max_budget,
                crash_cost=cls.crash_cost(metrics),
            )
            facade_cls = MultiFidelityFacade
        else:
            scenario = Scenario(
                configspace=space,
                seed=seed,
                output_directory=bucket.path / "smac3_output",
                deterministic=deterministic,
                objectives=metric_names,
                crash_cost=cls.crash_cost(metrics),
            )
            facade_cls = HyperparameterOptimizationFacade

        facade = facade_cls(
            scenario=scenario,
            target_function="dummy",  # NOTE: https://github.com/automl/SMAC3/issues/946
            overwrite=not continue_from_last_run,
            logging_level=logging_level,
            multi_objective_algorithm=facade_cls.get_multi_objective_algorithm(
                scenario=scenario,
            ),
        )
        return cls(facade=facade, fidelities=fidelities, bucket=bucket, metrics=metrics)

    @override
    def ask(self) -> Trial[SMACTrialInfo]:
        """Ask the optimizer for a new config.

        Returns:
            The trial info for the new config.
        """
        smac_trial_info = self.facade.ask()
        config = smac_trial_info.config
        budget = smac_trial_info.budget
        instance = smac_trial_info.instance
        seed = smac_trial_info.seed

        if self.fidelities and budget:
            if len(self.fidelities) == 1:
                k, _ = next(iter(self.fidelities.items()))
                trial_fids = {k: budget}
            else:
                trial_fids = {"budget": budget}
        else:
            trial_fids = None

        config_id = self.facade.runhistory.config_ids[config]
        unique_name = f"{config_id=}_{seed=}_{budget=}_{instance=}"
        trial: Trial[SMACTrialInfo] = Trial(
            name=unique_name,
            config=dict(config),
            info=smac_trial_info,
            seed=seed,
            fidelities=trial_fids,
            bucket=self.bucket,
            metrics=self.metrics,
        )
        logger.debug(f"Asked for trial {trial.name}")
        return trial

    @override
    def tell(self, report: Trial.Report[SMACTrialInfo]) -> None:
        """Tell the optimizer the result of the sampled config.

        Args:
            report: The report of the trial.
        """
        assert report.trial.info is not None

        cost: float | list[float]
        match self.metrics:
            case [metric]:  # Single obj
                val: Metric.Value = first_true(
                    report.metric_values,
                    pred=lambda m: m.metric == metric,
                    default=metric.worst,
                )
                cost = self.cost(val)
            case metrics:
                # NOTE: We need to make sure that there sorted in the order
                # that SMAC expects, with any missing metrics filled in
                _lookup = {v.metric.name: v for v in report.metric_values}
                cost = [
                    self.cost(_lookup.get(metric.name, metric.worst))
                    for metric in metrics
                ]

        logger.debug(f"Telling report for trial {report.trial.name}")

        # If we're successful, get the cost and times and report them
        params: dict[str, Any]
        match report.status:
            case Trial.Status.SUCCESS:
                params = {
                    "time": report.time.duration,
                    "starttime": report.time.start,
                    "endtime": report.time.end,
                    "cost": cost,
                    "status": StatusType.SUCCESS,
                }
            case Trial.Status.FAIL:
                params = {
                    "time": report.time.duration,
                    "starttime": report.time.start,
                    "endtime": report.time.end,
                    "cost": cost,
                    "status": StatusType.CRASHED,
                }
            case Trial.Status.CRASHED | Trial.Status.UNKNOWN:
                params = {
                    "cost": cost,
                    "status": StatusType.CRASHED,
                }

        match report.exception:
            case None:
                pass
            case MemoryLimitException():
                params["status"] = StatusType.MEMORYOUT
                params["additional_info"] = {
                    "exception": str(report.exception),
                    "traceback": report.traceback,
                }
            case TimeoutException():
                params["status"] = StatusType.TIMEOUT
                params["additional_info"] = {
                    "exception": str(report.exception),
                    "traceback": report.traceback,
                }
            case _:
                params["additional_info"] = {
                    "exception": str(report.exception),
                    "traceback": report.traceback,
                }

        self.facade.tell(report.trial.info, value=SMACTrialValue(**params), save=True)

    @override
    @classmethod
    def preferred_parser(cls) -> Literal["configspace"]:
        """The preferred parser for this optimizer."""
        return "configspace"

    @overload
    @classmethod
    def crash_cost(cls, metric: Metric) -> float:
        ...

    @overload
    @classmethod
    def crash_cost(cls, metric: Sequence[Metric]) -> list[float]:
        ...

    @classmethod
    def crash_cost(cls, metric: Metric | Sequence[Metric]) -> float | list[float]:
        """Get the crash cost for a metric for SMAC."""
        match metric:
            case Metric(bounds=(lower, upper)):  # Bounded metrics
                return abs(upper - lower)
            case Metric():  # Unbounded metric
                return np.inf
            case metrics:
                return [cls.crash_cost(m) for m in metrics]

    @classmethod
    def cost(cls, value: Metric.Value) -> float:
        """Get the cost for a metric value for SMAC."""
        match value.distance_to_optimal:
            case None:  # If we can't compute the distance, use the loss
                return value.loss
            case distance:  # If we can compute the distance, use that
                return distance



"""[ConfigSpace](https://automl.github.io/ConfigSpace/master/) is a library for
representing and sampling configurations for hyperparameter optimization.
It features a straightforward API for defining hyperparameters, their ranges
and even conditional dependencies.

It is generally flexible enough for more complex use cases, even
handling the complex pipelines of [AutoSklearn](https://automl.github.io/auto-sklearn/master/)
and [AutoPyTorch](https://automl.github.io/Auto-PyTorch/master/), large
scale hyperparameter spaces over which to optimize entire
pipelines at a time.

!!! tip "Requirements"

    This requires `ConfigSpace` which can be installed with:

    ```bash
    pip install "amltk[configspace]"

    # Or directly
    pip install ConfigSpace
    ```

In general, you should have the
[ConfigSpace documentation](https://automl.github.io/ConfigSpace/master/)
ready to consult for a full understanding of how to construct
hyperparameter spaces with AMLTK.

#### Basic Usage

You can directly us the [`parser()`][amltk.pipeline.parsers.configspace.parser]
function and pass that into the [`search_space()`][amltk.pipeline.Node.search_space]
method of a [`Node`][amltk.pipeline.Node], however you can also simply provide
`#!python search_space(parser="configspace", ...)` for simplicity.

```python exec="true" result="python" source="material-block" hl_lines="27" session="configspace-parser"
from amltk.pipeline import Component, Choice, Sequential
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC

my_pipeline = (
    Sequential(name="Pipeline")
    >> Component(PCA, space={"n_components": (1, 3)})
    >> Choice(
        Component(
            SVC,
            space={"C": (0.1, 10.0)}
        ),
        Component(
            RandomForestClassifier,
            space={"n_estimators": (10, 100), "criterion": ["gini", "log_loss"]},
        ),
        Component(
            MLPClassifier,
            space={
                "activation": ["identity", "logistic", "relu"],
                "alpha": (0.0001, 0.1),
                "learning_rate": ["constant", "invscaling", "adaptive"],
            },
        ),
        name="estimator"
    )
)

space = my_pipeline.search_space("configspace")
print(space)
```

Here we have an example of a few different kinds of hyperparmeters,

* `PCA:n_components` is a integer with a range of 1 to 3, uniform distribution, as specified
    by it's integer bounds in a tuple.
* `SVC:C` is a float with a range of 0.1 to 10.0, uniform distribution, as specified
    by it's float bounds in a tuple.
* `RandomForestClassifier:criterion` is a categorical hyperparameter, with two choices,
    `"gini"` and `"log_loss"`.

There is also a [`Choice`][amltk.pipeline.Choice] node, which is a special node that indicates that
we could choose from one of these estimators. This leads to the conditionals that you
can see in the printed out space.

You may wish to remove all conditionals if an `Optimizer` does not support them, or
you may wish to remove them for other reasons. You can do this by passing
`conditionals=False` to the [`parser()`][amltk.pipeline.parsers.configspace.parser] function.

```python exec="true" result="python" source="material-block" hl_lines="27" session="configspace-parser"
print(my_pipeline.search_space("configspace", conditionals=False))
```

Likewise, you can also remove all heirarchy from the space which may make downstream tasks easier,
by passing `flat=True` to the [`parser()`][amltk.pipeline.parsers.configspace.parser] function.

```python exec="true" result="python" source="material-block" hl_lines="27" session="configspace-parser"
print(my_pipeline.search_space("configspace", flat=True))
```

#### More Specific Hyperparameters
You'll often want to be a bit more specific with your hyperparameters, here we just
show a few examples of how you'd couple your pipelines a bit more towards `ConfigSpace`.

```python exec="true" result="python" source="material-block"
from ConfigSpace import Float, Categorical, Normal
from amltk.pipeline import Searchable

s = Searchable(
    space={
        "lr": Float("lr", bounds=(1e-5, 1.), log=True, default=0.3),
        "balance": Float("balance", bounds=(-1.0, 1.0), distribution=Normal(0.0, 0.5)),
        "color": Categorical("color", ["red", "green", "blue"], weights=[2, 1, 1], default="blue"),
    },
    name="Something-To-Search",
)
print(s.search_space("configspace"))
```

#### Conditional ands Advanced Usage
We will refer you to the
[ConfigSpace documentation](https://automl.github.io/ConfigSpace/master/) for the construction
of these. However once you've constructed a `ConfigurationSpace` and added any forbiddens and
conditionals, you may simply set that as the `.space` attribute.

```python exec="true" result="python" source="material-block" hl_lines="27"
from amltk.pipeline import Component, Choice, Sequential
from ConfigSpace import ConfigurationSpace, EqualsCondition, InCondition

myspace = ConfigurationSpace({"A": ["red", "green", "blue"], "B": (1, 10), "C": (-100.0, 0.0)})
myspace.add_conditions([
    EqualsCondition(myspace["B"], myspace["A"], "red"),  # B is active when A is red
    InCondition(myspace["C"], myspace["A"], ["green", "blue"]), # C is active when A is green or blue
])

component = Component(object, space=myspace, name="MyThing")

parsed_space = component.search_space("configspace")
print(parsed_space)
```

"""  # noqa: E501
from __future__ import annotations

from collections.abc import Mapping
from copy import deepcopy
from typing import Any

from ConfigSpace import Categorical, ConfigurationSpace, Constant

from amltk.pipeline import Choice, Node


def _remove_hyperparameter(
    name: str,
    space: ConfigurationSpace,
    seed: int | None = None,
) -> ConfigurationSpace:
    if name not in space._hyperparameters:
        raise ValueError(f"{name} not in {space}")

    # Copying conditionals only work on objects and not named entities
    # Seeing as we copy objects and don't use the originals, transfering these
    # to the new objects is a bit tedious, possible but not required at this time
    # ... same goes for forbiddens
    if name in space._conditionals:
        raise ValueError("Can't remove conditionals")
    if any(name == f.hyperparameter.name for f in space.get_forbiddens()):
        raise ValueError("Can't remove forbiddens")

    hps = [deepcopy(hp) for hp in space.get_hyperparameters() if hp.name != name]

    new_space = ConfigurationSpace(seed=seed, name=space.name, meta=space.meta)
    new_space.add_hyperparameters(hps)
    return new_space


def _remove_conditionals(
    space: ConfigurationSpace,
    seed: int | None = None,
) -> ConfigurationSpace:
    new_space = ConfigurationSpace(seed=seed, name=space.name, meta=space.meta)
    new_space.add_hyperparameters(space.values())
    return new_space


def _replace_constants(
    config: Mapping[str, Any],
    space: ConfigurationSpace,
    seed: int | None = None,
) -> ConfigurationSpace:
    for key, value in config.items():
        if key in space._hyperparameters:
            space = _remove_hyperparameter(key, space, seed)

            # These are just restrictions on hyperparameters from ConfigSpace
            match value:
                case bool():
                    space.add_hyperparameter(Constant(key, str(value)))
                case int() | float() | str():
                    space.add_hyperparameter(Constant(key, value))
                case _:
                    raise ValueError(f"Can't handle {value} from {config} as Constant")

    return space


def _parse_space(
    node: Node,
    *,
    conditionals: bool = True,
    seed: int | None = None,
) -> ConfigurationSpace:
    space = node.space
    match space:
        case ConfigurationSpace():
            _space = deepcopy(space)
        case Mapping():
            _space = ConfigurationSpace(dict(space))
        case None:
            _space = ConfigurationSpace()
        case _:
            raise ValueError(f"Can't handle {space} from {node}")

    if not conditionals:
        _space = _remove_conditionals(_space, seed)

    if node.config is not None:
        _space = _replace_constants(node.config, _space, seed)

    if seed is not None:
        _space.seed(seed)

    return _space


def parser(
    node: Node,
    *,
    seed: int | None = None,
    flat: bool = False,
    conditionals: bool = True,
    delim: str = ":",
) -> ConfigurationSpace:
    """Parse a Node and its children into a ConfigurationSpace.

    Args:
        node: The Node to parse
        seed: The seed to use for the ConfigurationSpace
        flat: Whether to have a heirarchical naming scheme for nodes and their children.
        conditionals: Whether to include conditionals in the space from a
            [`Choice`][amltk.pipeline.Choice]. If this is `False`, this will
            also remove all forbidden clauses and other conditional clauses.
            The primary use of this functionality is that some optimizers do not
            support these features.
        delim: The delimiter to use for the names of the hyperparameters
    """
    space = ConfigurationSpace(seed=seed)
    space.add_configuration_space(
        prefix=node.name,
        delimiter=delim,
        configuration_space=_parse_space(node, seed=seed, conditionals=conditionals),
    )

    children = node.nodes

    choice = None
    if isinstance(node, Choice) and any(children):
        choice = Categorical(
            name=f"{node.name}{delim}__choice__",
            items=[child.name for child in children],
        )
        space.add_hyperparameter(choice)

    for child in children:
        space.add_configuration_space(
            prefix=node.name if not flat else "",
            delimiter=delim if not flat else "",
            configuration_space=parser(
                child,
                seed=seed,
                flat=flat,
                conditionals=conditionals,
                delim=delim,
            ),
            parent_hyperparameter=(
                {"parent": choice, "value": child.name}
                if choice and conditionals
                else None
            ),
        )

    return space



'''A [`MetaFeature`][amltk.metalearning.MetaFeature] is some
statistic about a dataset/task, that can be used to make datasets or
tasks more comparable, thus enabling meta-learning methods.

Calculating meta-features of a dataset is quite straight foward.

```python exec="true" source="material-block" result="python" title="Metafeatures" hl_lines="10"
import openml
from amltk.metalearning import compute_metafeatures

dataset = openml.datasets.get_dataset(
    31,  # credit-g
    download_data=True,
    download_features_meta_data=False,
    download_qualities=False,
)
X, y, _, _ = dataset.get_data(
    dataset_format="dataframe",
    target=dataset.default_target_attribute,
)

mfs = compute_metafeatures(X, y)

print(mfs)
```

By default [`compute_metafeatures()`][amltk.metalearning.compute_metafeatures] will
calculate all the [`MetaFeature`][amltk.metalearning.MetaFeature] implemented,
iterating through their subclasses to do so. You can pass an explicit list
as well to `compute_metafeatures(X, y, features=[...])`.

To implement your own is also quite straight forward:

```python exec="true" source="material-block" result="python" title="Create Metafeature" hl_lines="10 11 12 13 14 15 16 17 18 19"
from amltk.metalearning import MetaFeature, compute_metafeatures
import openml

dataset = openml.datasets.get_dataset(
    31,  # credit-g
    download_data=True,
    download_features_meta_data=False,
    download_qualities=False,
)
X, y, _, _ = dataset.get_data(
    dataset_format="dataframe",
    target=dataset.default_target_attribute,
)

class TotalValues(MetaFeature):

    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: dict,
    ) -> int:
        return int(x.shape[0] * x.shape[1])

mfs = compute_metafeatures(X, y, features=[TotalValues])
print(mfs)
```

As many metafeatures rely on pre-computed dataset statistics, and they do not
need to be calculated more than once, you can specify the dependancies of
a meta feature. When a metafeature would return something other than a single
value, i.e. a `dict` or a `pd.DataFrame`, we instead call those a
[`DatasetStatistic`][amltk.metalearning.DatasetStatistic]. These will
**not** be included in the result of [`compute_metafeatures()`][amltk.metalearning.compute_metafeatures].
These `DatasetStatistic`s will only be calculated once on a call to `compute_metafeatures()` so
they can be re-used across all `MetaFeature`s that require that dependancy.

```python exec="true" source="material-block" result="python" title="Metafeature Dependancy" hl_lines="10 11 12 13 14 15 16 17 18 19 20 23 26 35"
from amltk.metalearning import MetaFeature, DatasetStatistic, compute_metafeatures
import openml

dataset = openml.datasets.get_dataset(
    31,  # credit-g
    download_data=True,
    download_features_meta_data=False,
    download_qualities=False,
)
X, y, _, _ = dataset.get_data(
    dataset_format="dataframe",
    target=dataset.default_target_attribute,
)

class NAValues(DatasetStatistic):
    """A mask of all NA values in a dataset"""

    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: dict,
    ) -> pd.DataFrame:
        return x.isna()


class PercentageNA(MetaFeature):
    """The percentage of values missing"""

    dependencies = (NAValues,)

    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: dict,
    ) -> int:
        na_values = dependancy_values[NAValues]
        n_na = na_values.sum().sum()
        n_values = int(x.shape[0] * x.shape[1])
        return float(n_na / n_values)

mfs = compute_metafeatures(X, y, features=[PercentageNA])
print(mfs)
```

To view the description of a particular `MetaFeature`, you can call
[`.description()`][amltk.metalearning.DatasetStatistic.description]
on it. Otherwise you can access all of them in the following way:

```python exec="true" source="tabbed-left" result="python" title="Metafeature Descriptions" hl_lines="4"
from pprint import pprint
from amltk.metalearning import metafeature_descriptions

descriptions = metafeature_descriptions()
for name, description in descriptions.items():
    print("---")
    print(name)
    print("---")
    print(" * " + description)
```
'''  # noqa: E501
from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, Mapping
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeAlias, TypeVar
from typing_extensions import override

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

CAMEL_CASE_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")

if TYPE_CHECKING:
    DSdict: TypeAlias = dict[type["DatasetStatistic"], Any]

S = TypeVar("S")
M = TypeVar("M", bound=float)
T = TypeVar("T")

KURTOSIS_OF_NORMAL_DISTRIBUTION = 0.0
SKEWNESS_OF_NORMAL_DISTRIBUTION = 0.0


def imbalance_ratios(col: pd.Series | pd.DataFrame) -> tuple[pd.Series, float]:
    """Compute the imbalance ratio of a categorical column.

    This is done by computing the distance of each item's ratio to what
    a perfectly balanced ratio would be. We then sum up the distances,
    dividing by the worst case to normalize between 0 and 1.

    Args:
        col: A column of values. If a DataFrame, the values from the subset of columns
            will be used.

    Returns:
        A tuple of the imbalance ratios, sorted from lowest (0) to highest (1)
        and the expected ratio if perfectly balanced.
    """
    ratios = col.value_counts(dropna=True, normalize=True, ascending=True)
    if len(ratios) == 1:
        return ratios, 1.0

    n_uniq = len(ratios)

    # A balanced ratio is one where all items are equally distributed
    balanced_ratio = float(1 / n_uniq)
    return ratios, balanced_ratio


def column_imbalance(ratios: pd.Series, balanced_ratio: float) -> float:
    """Compute the imbalance of a column.

    This is done by computing the distance of each item's ratio to what
    a perfectly balanced ratio would be. We then sum up the distances,
    dividing by the worst case to normalize between 0 and 1. 0 indicates
    a perfectly balanced column, 1 indicates a column where all items
    are of the same type.

    Args:
        ratios: The ratios of each item in the column.
        balanced_ratio: The ratio of a column if perfectly balanced.

    Returns:
        The imbalance of the column.
    """
    item_ratios_distance_from_balanced_ratio = np.abs(ratios - balanced_ratio)

    # The most imbalanced dataset would be one where we somehow have 0
    # items of each type **except** 1 type, which has all the instances.

    # In the case of a symbol group with 0 instance, their distance to the balanced
    # ratio is just the balanced ratio itself.
    zero_instance_ratio_distance = balanced_ratio
    dominant_ratio_distance = np.abs(1 - balanced_ratio)
    n_items = len(ratios)

    worst = (n_items - 1) * zero_instance_ratio_distance + dominant_ratio_distance
    normalizer = 1 / worst

    return float(normalizer * np.sum(item_ratios_distance_from_balanced_ratio))


class DatasetStatistic(ABC, Generic[S]):
    """Base class for a dataset statistic.

    A dataset statistic is a function that takes a dataset and returns some
    value(s) that describe the dataset.

    If looking to create meta-features, see the `MetaFeature` class which
    restricts the statistic to be a single number.
    """

    dependencies: ClassVar[tuple[type[DatasetStatistic], ...]] = ()

    @classmethod
    def description(cls) -> str:
        """Return the description of this statistic."""
        return cls.__doc__ or ""

    @classmethod
    def name(cls) -> str:
        """Return the name of this statistic."""
        return CAMEL_CASE_PATTERN.sub("_", cls.__name__).lower()

    @classmethod
    @abstractmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> S:
        """Compute the value of this statistic.

        Args:
            x: The features of the dataset.
            y: The labels of the dataset.
            dependancy_values: A dictionary of dependency values.

        Returns:
            The value of this statistic.
        """

    @classmethod
    def retrieve(
        cls,
        dependancy_values: Mapping[type[DatasetStatistic[T]], T],
    ) -> S:
        """Retrieve the value of this statistic from the dependency values.

        Args:
            dependancy_values: A dictionary of dependency values.

        Returns:
            The value of this statistic.
        """
        return dependancy_values[cls]  # type: ignore


class MetaFeature(DatasetStatistic[M]):
    """Used to indicate a metafeature to include.

    This differs from DatasetStatistic in that it must return a single value.
    """

    skip: ClassVar[bool] = False
    """Whether to skip this metafeature when
    [`iter()`][amltk.metalearning.MetaFeature.iter] is called.
    """

    @classmethod
    def iter(cls) -> Iterator[type[MetaFeature]]:
        """Return all the subclasses of MetaFeature."""
        for c in cls.__subclasses__():
            if not c.skip:
                yield c


class NAValues(DatasetStatistic[pd.DataFrame]):
    """Mask of missing values in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> pd.DataFrame:
        return x.isna()


class ClassImbalanceRatios(DatasetStatistic[tuple[pd.Series, float]]):
    """Imbalance ratios of each class in the dataset.

    Will return the ratios of each class, the ratio expected if perfectly balanced,
    """

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> tuple[pd.Series, float]:
        return imbalance_ratios(y)


class CategoricalImbalanceRatios(DatasetStatistic[dict[str, tuple[pd.Series, float]]]):
    """Imbalance ratios of each class in the dataset.

    Will return the ratios of each class, the ratio expected if perfectly balanced,
    """

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> dict[str, tuple[pd.Series, float]]:
        categorical_columns = x.select_dtypes(exclude=np.number).columns
        return {c: imbalance_ratios(x[c]) for c in categorical_columns}


class CategoricalColumns(DatasetStatistic[pd.DataFrame]):
    """The categorical columns in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> pd.DataFrame:
        return x.select_dtypes(exclude=np.number)


class NumericalColumns(DatasetStatistic[pd.DataFrame]):
    """The numerical columns in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> pd.DataFrame:
        return x.select_dtypes(include=np.number)


class InstanceCount(MetaFeature[int]):
    """Number of instances in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> int:
        return int(x.shape[0])


class LogInstanceCount(MetaFeature[float]):
    """Logarithm of the number of instances in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        return float(np.log(x.shape[0]))


class NumberOfClasses(MetaFeature[int]):
    """Number of classes in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> int:
        if len(y.shape) == 1:
            return len(y.value_counts(sort=False))

        # NOTE: Not really sure "how many" classes
        # there are in the case of multi-label classification
        # I guess this meta-feature doesn't make sense then?
        if len(y.shape) == 2:  # noqa: PLR2004
            return int(y.shape[1])

        raise ValueError("y must be 1D or 2D")


class NumberOfFeatures(MetaFeature[int]):
    """Number of features in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> int:
        return int(x.shape[1])


class LogNumberOfFeatures(MetaFeature[float]):
    """Logarithm of the number of features in the dataset."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        return float(np.log(x.shape[1]))


class PercentageMissingValues(MetaFeature[float]):
    """Percentage of missing values in the dataset."""

    dependencies = (NAValues,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_values = x.shape[0] * x.shape[1]
        na_values = NAValues.retrieve(dependancy_values)
        n_missing = na_values.sum().sum()
        return float(n_missing / n_values)


class PercentageOfInstancesWithMissingValues(MetaFeature[float]):
    """Percentage of instances with missing values."""

    dependencies = (NAValues,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_instances = x.shape[0]
        na_values = NAValues.retrieve(dependancy_values)
        n_instances_with_missing_values = na_values.any(axis=1).sum()
        return float(n_instances_with_missing_values / n_instances)


class PercentageOfFeaturesWithMissingValues(MetaFeature[float]):
    """Percentage of features with missing values."""

    dependencies = (NAValues,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_features = x.shape[1]
        na_values = NAValues.retrieve(dependancy_values)
        n_features_with_missing_values = na_values.any(axis=0).sum()
        return float(n_features_with_missing_values / n_features)


class PercentageOfCategoricalColumnsWithMissingValues(MetaFeature[float]):
    """Percentage of categorical columns with missing values."""

    dependencies = (CategoricalColumns, NAValues)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        categorical_columns = CategoricalColumns.retrieve(dependancy_values)
        if categorical_columns.empty:
            return 0.0

        na_values = NAValues.retrieve(dependancy_values)
        n_categorical_features = categorical_columns.shape[1]
        n_categorical_features_with_missing_values = (
            na_values[categorical_columns.columns].any(axis=0).sum()
        )
        return float(
            n_categorical_features_with_missing_values / n_categorical_features,
        )


class PercentageOfCategoricalValuesWithMissingValues(MetaFeature[float]):
    """Percentage of categorical values with missing values."""

    dependencies = (CategoricalColumns, NAValues)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        categorical_columns = CategoricalColumns.retrieve(dependancy_values)
        if categorical_columns.empty:
            return 0.0

        na_values = NAValues.retrieve(dependancy_values)
        n_categorical_values = (
            categorical_columns.shape[0] * categorical_columns.shape[1]
        )
        n_categorical_values_with_missing_values = (
            na_values[categorical_columns.columns].sum().sum()
        )
        return float(n_categorical_values_with_missing_values / n_categorical_values)


class PercentageOfNumericColumnsWithMissingValues(MetaFeature[float]):
    """Percentage of numeric columns with missing values."""

    dependencies = (NumericalColumns, NAValues)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        numerical_columns = NumericalColumns.retrieve(dependancy_values)
        if numerical_columns.empty:
            return 0.0

        na_values = NAValues.retrieve(dependancy_values)
        n_numerical_features = numerical_columns.shape[1]
        n_numerical_features_with_missing_values = (
            na_values[numerical_columns.columns].any(axis=0).sum()
        )
        return float(n_numerical_features_with_missing_values / n_numerical_features)


class PercentageOfNumericValuesWithMissingValues(MetaFeature[float]):
    """Percentage of numeric values with missing values."""

    dependencies = (NumericalColumns, NAValues)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        numerical_columns = NumericalColumns.retrieve(dependancy_values)
        if numerical_columns.empty:
            return 0.0

        na_values = NAValues.retrieve(dependancy_values)
        n_numerical_values = numerical_columns.shape[0] * numerical_columns.shape[1]
        n_numerical_values_with_missing_values = (
            na_values[numerical_columns.columns].sum().sum()
        )
        return float(n_numerical_values_with_missing_values / n_numerical_values)


class NumberOfNumericFeatures(MetaFeature[int]):
    """Number of numeric features in the dataset."""

    dependencies = (NumericalColumns,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> int:
        numerical_columns = NumericalColumns.retrieve(dependancy_values)
        if numerical_columns.empty:
            return 0

        return int(numerical_columns.shape[1])


class NumberOfCategoricalFeatures(MetaFeature[int]):
    """Number of categorical features in the dataset."""

    dependencies = (CategoricalColumns,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> int:
        categorical_columns = CategoricalColumns.retrieve(dependancy_values)
        if categorical_columns.empty:
            return 0

        return int(categorical_columns.shape[1])


class RatioNumericalFeatures(MetaFeature[float]):
    """Ratio of numerical features to total features in the dataset."""

    dependencies = (NumberOfNumericFeatures,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_numerical = NumberOfNumericFeatures.retrieve(dependancy_values)
        return float(n_numerical / x.shape[1])


class RatioCategoricalFeatures(MetaFeature[float]):
    """Ratio of categoricals features to total features in the dataset."""

    dependencies = (NumberOfCategoricalFeatures,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_categorical = NumberOfCategoricalFeatures.retrieve(dependancy_values)
        return float(n_categorical / x.shape[1])


class RatioFeaturesToInstances(MetaFeature[float]):
    """Ratio of features to instances in the dataset."""

    dependencies = (NumberOfFeatures, InstanceCount)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        n_features = NumberOfFeatures.retrieve(dependancy_values)
        n_instances = InstanceCount.retrieve(dependancy_values)
        return float(n_features / n_instances)


class ClassCounts(DatasetStatistic[pd.Series]):
    """Number of instances per class."""

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> pd.Series:
        return y.value_counts()


class MinorityClassImbalance(MetaFeature[float]):
    """Imbalance of the minority class in the dataset. 0 => Balanced. 1 imbalanced."""

    dependencies = (ClassImbalanceRatios,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        ratios, balanced_ratio = ClassImbalanceRatios.retrieve(dependancy_values)
        minority_ratio = ratios.iloc[0]
        distance_to_balanced = np.abs(minority_ratio - balanced_ratio)
        # This happens when all the ratio is in one class
        worst_case = (len(ratios) - 1) * balanced_ratio + (1 - balanced_ratio)
        return float(distance_to_balanced / worst_case)


class MajorityClassImbalance(MetaFeature[float]):
    """Imbalance of the majority class in the dataset. 0 => Balanced. 1 imbalanced."""

    dependencies = (ClassImbalanceRatios,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        ratios, balanced_ratio = ClassImbalanceRatios.retrieve(dependancy_values)
        majority_ratio = ratios.iloc[-1]
        distance_to_balanced = np.abs(majority_ratio - balanced_ratio)
        # This happens when all the ratio is in one class
        worst_case = (len(ratios) - 1) * balanced_ratio + (1 - balanced_ratio)
        return float(distance_to_balanced / worst_case)


class ClassImbalance(MetaFeature[float]):
    """Mean Target Imbalance of the classes in general.

    0 => Balanced. 1 Imbalanced.
    """

    dependencies = (ClassImbalanceRatios,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        ratios, balanced_ratio = ClassImbalanceRatios.retrieve(dependancy_values)
        return column_imbalance(ratios, balanced_ratio)


class ImbalancePerCategory(DatasetStatistic[dict[str, float]]):
    """Imbalance of each categorical feature. 0 => Balanced. 1 most imbalanced.

    No categories implies perfectly balanced.
    """

    dependencies = ()

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> dict[str, float]:
        categoricals = x.select_dtypes(exclude=np.number)
        if categoricals.empty:
            return {}

        return {
            str(col): column_imbalance(*imbalance_ratios(x[col]))
            for col in categoricals.columns
        }


class MeanCategoricalImbalance(MetaFeature[float]):
    """The mean imbalance of categorical features."""

    dependencies = (ImbalancePerCategory,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        imbalances = ImbalancePerCategory.retrieve(dependancy_values)
        if len(imbalances) == 0:
            return 0

        return float(np.mean(list(imbalances.values())))


class StdCategoricalImbalance(MetaFeature[float]):
    """The std imbalance of categorical features."""

    dependencies = (ImbalancePerCategory,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        imbalances = ImbalancePerCategory.retrieve(dependancy_values)
        if len(imbalances) == 0:
            return 0

        return float(np.std(list(imbalances.values())))


class SkewnessPerNumericalColumn(DatasetStatistic[dict[str, float]]):
    """Skewness of each numerical feature."""

    dependencies = (NumericalColumns,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> dict[str, float]:
        numericals = NumericalColumns.retrieve(dependancy_values)
        if numericals.empty:
            return {}

        skews: pd.Series = x.skew(numeric_only=True)  # type: ignore
        return dict(skews)


class SkewnessMean(MetaFeature[float]):
    """The mean skewness of numerical features."""

    dependencies = (SkewnessPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        skews = SkewnessPerNumericalColumn.retrieve(dependancy_values)
        if len(skews) == 0:
            return SKEWNESS_OF_NORMAL_DISTRIBUTION

        return float(np.mean(list(skews.values())))


class SkewnessStd(MetaFeature[float]):
    """The std skewness of numerical features."""

    dependencies = (SkewnessPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        skews = SkewnessPerNumericalColumn.retrieve(dependancy_values)
        if len(skews) == 0:
            return SKEWNESS_OF_NORMAL_DISTRIBUTION

        return float(np.std(list(skews.values())))


class SkewnessMin(MetaFeature[float]):
    """The min skewness of numerical features."""

    dependencies = (SkewnessPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        skews = SkewnessPerNumericalColumn.retrieve(dependancy_values)
        if len(skews) == 0:
            return SKEWNESS_OF_NORMAL_DISTRIBUTION

        return float(np.min(list(skews.values())))


class SkewnessMax(MetaFeature[float]):
    """The max skewness of numerical features."""

    dependencies = (SkewnessPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        skews = SkewnessPerNumericalColumn.retrieve(dependancy_values)
        if len(skews) == 0:
            return SKEWNESS_OF_NORMAL_DISTRIBUTION

        return float(np.max(list(skews.values())))


class KurtosisPerNumericalColumn(DatasetStatistic[dict[str, float]]):
    """Kurtosis of each numerical feature."""

    dependencies = (NumericalColumns,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> dict[str, float]:
        numericals = NumericalColumns.retrieve(dependancy_values)
        if numericals.empty:
            return {}

        kurts: pd.Series = numericals.kurt()  # type: ignore
        return dict(kurts)


class KurtosisMean(MetaFeature[float]):
    """The mean kurtosis of numerical features."""

    dependencies = (KurtosisPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        kurts = KurtosisPerNumericalColumn.retrieve(dependancy_values)
        if len(kurts) == 0:
            return KURTOSIS_OF_NORMAL_DISTRIBUTION

        return float(np.mean(list(kurts.values())))


class KurtosisStd(MetaFeature[float]):
    """The std kurtosis of numerical features."""

    dependencies = (KurtosisPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        kurts = KurtosisPerNumericalColumn.retrieve(dependancy_values)
        if len(kurts) == 0:
            return KURTOSIS_OF_NORMAL_DISTRIBUTION

        return float(np.std(list(kurts.values())))


class KurtosisMin(MetaFeature[float]):
    """The min kurtosis of numerical features."""

    dependencies = (KurtosisPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        kurts = KurtosisPerNumericalColumn.retrieve(dependancy_values)
        if len(kurts) == 0:
            return KURTOSIS_OF_NORMAL_DISTRIBUTION

        return float(np.min(list(kurts.values())))


class KurtosisMax(MetaFeature[float]):
    """The max kurtosis of numerical features."""

    dependencies = (KurtosisPerNumericalColumn,)

    @override
    @classmethod
    def compute(
        cls,
        x: pd.DataFrame,
        y: pd.Series | pd.DataFrame,
        dependancy_values: DSdict,
    ) -> float:
        kurts = KurtosisPerNumericalColumn.retrieve(dependancy_values)
        if len(kurts) == 0:
            return KURTOSIS_OF_NORMAL_DISTRIBUTION

        return float(np.max(list(kurts.values())))


def metafeature_descriptions(
    features: Iterable[type[DatasetStatistic]] | None = None,
) -> dict[str, str]:
    """Get the descriptions of meatfeatures available.

    Args:
        features: The metafeatures. If None, all metafeatures subclasses
            of [`MetaFeature`][amltk.metalearning.MetaFeature] will be returned.

    Returns:
        The descriptions of the metafeatures.
    """
    if features is None:
        features = MetaFeature.iter()

    return {mf.name(): mf.description() for mf in features}


def compute_metafeatures(
    X: pd.DataFrame,  # noqa: N803
    y: pd.Series | pd.DataFrame,
    *,
    features: Iterable[type[MetaFeature]] | None = None,
) -> pd.Series:
    """Compute metafeatures for a dataset.

    Args:
        X: The features of the dataset.
        y: The labels of the dataset.
        features: The metafeatures to compute. If None, all metafeatures subclasses
            of [`MetaFeature`][amltk.metalearning.MetaFeature] will be computed.

    Returns:
        A series of metafeatures.
    """
    if features is None:
        features = MetaFeature.iter()

    def _calc(
        _x: pd.DataFrame,
        _y: pd.Series | pd.DataFrame,
        _metafeature: type[DatasetStatistic],
        _values: dict[type[DatasetStatistic], Any],
    ) -> dict[type[DatasetStatistic], Any]:
        for dep in _metafeature.dependencies:
            _values = _calc(_x, _y, dep, _values)

        if _metafeature not in _values:
            _values[_metafeature] = _metafeature.compute(_x, _y, _values)

        return _values

    values: dict[type[DatasetStatistic], Any] = {}
    for mf in features:
        values = _calc(X, y, mf, values)

    return pd.Series(
        {
            key.name(): value
            for key, value in values.items()
            if issubclass(key, MetaFeature)
        },
    )
